# pylint: disable=missing-function-docstring
"""
Environment variable and settings management. See the README for information
about the supported environment variables.
"""
__docformat__ = "google"

import logging
import os
from pathlib import Path
from typing import Any, Dict, Union

_DATACLASSES_TYPES: Dict[str, type] = {}

# The initial values are the defaults
_ENVIRONMENT: Dict[str, Any] = {
    "TB_ARTIFACT_PATH": Path("./"),
    "TB_KERAS_FORMAT": "tf",
    "TB_PANDAS_FORMAT": "h5",
    "TB_MAX_NBYTES": 8_000,
}


def _init():
    """
    Reads the environment and sets the
    `turbo_broccoli.environment._ENVIRONMENT` accordingly.
    """
    if "TB_NUMPY_PATH" in os.environ:
        logging.warning(
            "The use of the TB_NUMPY_PATH environment variable is deprecated. "
            "Consider using TB_ARTIFACT_PATH instead"
        )
        set_artifact_path(Path(os.environ["TB_NUMPY_PATH"]))
    else:
        set_artifact_path(
            os.environ.get(
                "TB_ARTIFACT_PATH",
                _ENVIRONMENT["TB_ARTIFACT_PATH"],
            )
        )

    try:
        set_keras_format(
            os.environ.get(
                "TB_KERAS_FORMAT",
                _ENVIRONMENT["TB_KERAS_FORMAT"],
            )
        )
    except ValueError:
        logging.warning(
            "Invalid value for environment variable TB_KERAS_FORMAT: '%s'. "
            "Expected 'h5', 'json', or 'tf'. Defaulting to '%s'",
            os.environ["TB_KERAS_FORMAT"],
            _ENVIRONMENT["TB_KERAS_FORMAT"],
        )

    try:
        set_pandas_format(
            os.environ.get(
                "TB_PANDAS_FORMAT",
                _ENVIRONMENT["TB_PANDAS_FORMAT"],
            )
        )
    except ValueError:
        logging.warning(
            "Invalid value for environment variable TB_PANDAS_FORMAT: '%s'. "
            "Expected 'csv', 'excel', 'feather', 'h5', 'hdf', 'pickle', "
            "'stata', or 'xml'. Defaulting to '%s'",
            os.environ["TB_PANDAS_FORMAT"],
            _ENVIRONMENT["TB_PANDAS_FORMAT"],
        )

    if "TB_NUMPY_MAX_NBYTES" in os.environ:
        logging.warning(
            "The use of the TB_NUMPY_MAX_NBYTES environment variable is "
            "deprecated. Consider using TB_MAX_NBYTES instead"
        )
        set_max_nbytes(int(os.environ["TB_NUMPY_MAX_NBYTES"]))
    else:
        set_max_nbytes(
            int(
                os.environ.get(
                    "TB_MAX_NBYTES",
                    _ENVIRONMENT["TB_MAX_NBYTES"],
                )
            )
        )


def get_artifact_path() -> Path:
    return _ENVIRONMENT["TB_ARTIFACT_PATH"]


def get_keras_format() -> str:
    return _ENVIRONMENT["TB_KERAS_FORMAT"]


def get_pandas_format() -> str:
    return _ENVIRONMENT["TB_PANDAS_FORMAT"]


def get_max_nbytes() -> int:
    return _ENVIRONMENT["TB_MAX_NBYTES"]


def get_register_dataclass(name: str) -> type:
    return _DATACLASSES_TYPES[name]


def register_dataclass(name: str, cls: type):
    """
    Registers a dataclass for dataclass deserialization. Registered types may
    be overwritten.
    """
    _DATACLASSES_TYPES[name] = cls


def set_artifact_path(path: Union[str, Path]):
    if isinstance(path, str):
        path = Path(path)
    if not (path.exists() and path.is_dir()):
        raise RuntimeError(
            f"Path {str(path)} does not point to an existing directory"
        )
    _ENVIRONMENT["TB_ARTIFACT_PATH"] = path


def set_keras_format(fmt: str):
    fmt = fmt.lower()
    KERAS_FORMATS = ["h5", "json", "tf"]
    if fmt not in KERAS_FORMATS:
        raise ValueError(
            f"Invalid value for environment variable TB_KERAS_FORMAT: {fmt}."
        )
    _ENVIRONMENT["TB_KERAS_FORMAT"] = fmt


def set_pandas_format(fmt: str):
    fmt = fmt.lower()
    PANDAS_FORMATS = [
        "csv",
        "excel",
        "feather",
        "h5",
        "hdf",
        "html",
        "pickle",
        "sql",
        "stata",
        "xml",
    ]
    if fmt not in PANDAS_FORMATS:
        raise ValueError(
            f"Invalid value for environment variable TB_PANDAS_FORMAT: {fmt}."
        )
    _ENVIRONMENT["TB_PANDAS_FORMAT"] = fmt


def set_max_nbytes(nbytes: int):
    if nbytes <= 0:
        raise ValueError("numpy's max nbytes must be > 0")
    _ENVIRONMENT["TB_MAX_NBYTES"] = nbytes


_init()



"""
.. include:: ../README.md
.. include:: ../CHANGELOG.md
"""
__docformat__ = "google"

from pkg_resources import get_distribution, DistributionNotFound

from turbo_broccoli.turbo_broccoli import (
    TurboBroccoliDecoder,
    TurboBroccoliEncoder,
)

try:
    __version__ = get_distribution("turbo-broccoli").version
except DistributionNotFound:
    __version__ = "local"



"""
numpy (de)serialization utilities.

Todo:
    Handle numpy's `generic` type (which supersedes the `number` type).
"""
__docformat__ = "google"

from typing import Any, Callable, List, Tuple
from uuid import uuid4

import numpy as np

from turbo_broccoli.environment import get_artifact_path, get_max_nbytes


def _json_to_ndarray(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_ndarray_v1,
        2: _json_to_ndarray_v2,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_ndarray_v1(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array following the v1 specification.
    """
    return np.frombuffer(
        dct["data"],
        dtype=np.lib.format.descr_to_dtype(dct["dtype"]),
    ).reshape(dct["shape"])


def _json_to_ndarray_v2(dct: dict) -> np.ndarray:
    """
    Converts a JSON document to a numpy array following the v1 specification.
    """
    if "data" in dct:
        return _json_to_ndarray_v1(dct)
    return np.load(get_artifact_path() / (dct["id"] + ".npy"))


def _json_to_number(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy number. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__numpy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_number_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_number_v1(dct: dict) -> np.number:
    """
    Converts a JSON document to a numpy number following the v1 specification.
    """
    return np.frombuffer(
        dct["value"],
        dtype=np.lib.format.descr_to_dtype(dct["dtype"]),
    )[0]


def _ndarray_to_json(arr: np.ndarray) -> dict:
    """Serializes a `numpy` array."""
    if arr.nbytes <= get_max_nbytes():
        return {
            "__type__": "ndarray",
            "__version__": 2,
            "data": bytes(arr.data),
            "dtype": np.lib.format.dtype_to_descr(arr.dtype),
            "shape": arr.shape,
        }
    name = str(uuid4())
    np.save(get_artifact_path() / name, arr)
    return {
        "__type__": "ndarray",
        "__version__": 2,
        "id": name,
    }


def _number_to_json(num: np.number) -> dict:
    """Serializes a `numpy` number."""

    return {
        "__type__": "number",
        "__version__": 1,
        "value": bytes(np.array(num).data),
        "dtype": np.lib.format.dtype_to_descr(num.dtype),
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a numpy object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__numpy__`.
    """
    DECODERS = {
        "ndarray": _json_to_ndarray,
        "number": _json_to_number,
    }
    try:
        return DECODERS[dct["__numpy__"]["__type__"]](dct["__numpy__"])
    except KeyError as exc:
        raise TypeError("Not a valid numpy document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a `numpy` object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__numpy__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    * `numpy.ndarray`: An array is processed differently depending on its size
      and on the `TB_MAX_NBYTES` environment variable. If the array is
      small, i.e. `arr.nbytes <= TB_MAX_NBYTES`, then it is directly
      stored in the resulting JSON document as

            {
                "__numpy__": {
                    "__type__": "ndarray",
                    "__version__": 2,
                    "data": <ASCII encoded byte string>,
                    "dtype": <str>,
                    "shape": <int tuple>,
                }
            }

      On the other hand, if `arr.nbytes > TB_MAX_NBYTES`, then the
      content of `arr` is stored in an `.npy` file. Said file is saved to the
      path specified by the `TB_ARTIFACT_PATH` environment variable with a
      random UUID4 as filename. The resulting JSON document looks like

            {
                "__numpy__": {
                    "__type__": "ndarray",
                    "__version__": 2,
                    "id": <UUID4 str>,
                }
            }

      By default, `TB_MAX_NBYTES` is `8000` bytes, which should be enough
      to store an array of 1000 `float64`s, and `TB_ARTIFACT_PATH` is `./`.
      `TB_ARTIFACT_PATH` must point to an existing directory.

    * `numpy.number`:

            {
                "__numpy__": {
                    "__type__": "number",
                    "__version__": 1,
                    "value": <float>,
                    "dtype": <str>,
                }
            }

    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (np.ndarray, _ndarray_to_json),
        (np.number, _number_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__numpy__": f(obj)}
    raise TypeError("Not a supported numpy type")



"""Tensorflow (de)serialization utilities."""
__docformat__ = "google"

from typing import Any, Callable, List, Tuple

import tensorflow as tf


def _json_to_sparse_tensor(dct: dict) -> tf.Tensor:
    """Converts a JSON document to a tensorflow tensor."""
    DECODERS = {
        1: _json_to_sparse_tensor_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_sparse_tensor_v1(dct: dict) -> tf.Tensor:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    sparse tensor.
    """
    return tf.SparseTensor(
        dense_shape=dct["shape"],
        indices=dct["indices"],
        values=dct["values"],
    )


def _json_to_tensor(dct: dict) -> tf.Tensor:
    """Converts a JSON document to a tensorflow tensor."""
    DECODERS = {
        1: _json_to_tensor_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_tensor_v1(dct: dict) -> tf.Tensor:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    tensor.
    """
    return tf.constant(dct["numpy"], dtype=dct["dtype"])


def _json_to_variable(dct: dict) -> tf.Variable:
    """Converts a JSON document to a tensorflow variable."""
    DECODERS = {
        1: _json_to_variable_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_variable_v1(dct: dict) -> tf.Variable:
    """
    Converts a JSON document following the v1 specification to a tensorflow
    variable.
    """
    return tf.Variable(
        dtype=dct["dtype"],
        initial_value=dct["numpy"],
        name=dct["name"],
        trainable=dct["trainable"],
    )


def _ragged_tensor_to_json(tens: tf.Tensor) -> dict:
    """Serializes a general tensor"""
    raise NotImplementedError(
        "Serialization of ragged tensors is not supported"
    )


def _sparse_tensor_to_json(tens: tf.SparseTensor) -> dict:
    """Serializes a sparse tensor"""
    return {
        "__type__": "sparse_tensor",
        "__version__": 1,
        "indices": tens.indices,
        "shape": tens.dense_shape.numpy(),
        "values": tens.values,
    }


def _tensor_to_json(tens: tf.Tensor) -> dict:
    """Serializes a general tensor"""
    return {
        "__type__": "tensor",
        "__version__": 1,
        "dtype": tens.dtype.name,
        "numpy": tens.numpy(),
    }


def _variable_to_json(var: tf.Variable) -> dict:
    """Serializes a general tensor"""
    return {
        "__type__": "variable",
        "__version__": 1,
        "dtype": var.dtype.name,
        "name": var.name,
        "numpy": var.numpy(),
        "trainable": var.trainable,
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a tensorflow object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__tensorflow__`.
    """
    DECODERS = {
        "sparse_tensor": _json_to_sparse_tensor,
        "tensor": _json_to_tensor,
        "variable": _json_to_variable,
    }
    try:
        return DECODERS[dct["__tensorflow__"]["__type__"]](
            dct["__tensorflow__"]
        )
    except KeyError as exc:
        raise TypeError("Not a valid tensorflow document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a tensorflow object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__tensorflow__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    * `tf.RaggedTensor`: Not supported.
    * `tf.SparseTensor`:

            {
                "__type__": "sparse_tensor",
                "__version__": 1,
                "indices": {...},
                "values": {...},
                "shape": {...},
            }

      where the first two `{...}` placeholders result in the serialization of
      `tf.Tensor` (see below).
    * other `tf.Tensor` subtypes:

            {
                "__type__": "tensor",
                "__version__": 1,
                "dtype": <str>,
                "numpy": {...},
            }

      where `{...}` is the document produced by `turbo_broccoli.numpy.to_json`.
    * `tf.Variable`:

            {
                "__type__": "tensor",
                "__version__": 1,
                "dtype": <str>,
                "name": <str>,
                "numpy": {...},
                "trainable": <bool>,
            }

      where `{...}` is the document produced by `turbo_broccoli.numpy.to_json`.

    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (tf.RaggedTensor, _ragged_tensor_to_json),
        (tf.SparseTensor, _sparse_tensor_to_json),
        (tf.Tensor, _tensor_to_json),
        (tf.Variable, _variable_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__tensorflow__": f(obj)}
    raise TypeError("Not a supported tensorflow type")



"""Main module containing the JSON encoder and decoder classes."""
__docformat__ = "google"

import json
from typing import Any, Callable, Dict, List


import turbo_broccoli.bytes
import turbo_broccoli.collections
import turbo_broccoli.dataclass
import turbo_broccoli.generic

try:
    import turbo_broccoli.keras

    HAS_KERAS = True
except ModuleNotFoundError:
    HAS_KERAS = False

try:
    import turbo_broccoli.numpy

    HAS_NUMPY = True
except ModuleNotFoundError:
    HAS_NUMPY = False

try:
    import turbo_broccoli.pandas

    HAS_PANDAS = True
except ModuleNotFoundError:
    HAS_PANDAS = False

try:
    import turbo_broccoli.tensorflow

    HAS_TENSORFLOW = True
except ModuleNotFoundError:
    HAS_TENSORFLOW = False


class TurboBroccoliDecoder(json.JSONDecoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(object_hook=self._hook, *args, **kwargs)

    def _hook(self, dct):
        """Deserialization hook"""
        DECODERS: Dict[str, Callable[[dict], Any]] = {
            "__bytes__": turbo_broccoli.bytes.from_json,
            "__collections__": turbo_broccoli.collections.from_json,
            "__dataclass__": turbo_broccoli.dataclass.from_json,
        }
        if HAS_KERAS:
            DECODERS["__keras__"] = turbo_broccoli.keras.from_json
        if HAS_NUMPY:
            DECODERS["__numpy__"] = turbo_broccoli.numpy.from_json
        if HAS_PANDAS:
            DECODERS["__pandas__"] = turbo_broccoli.pandas.from_json
        if HAS_TENSORFLOW:
            DECODERS["__tensorflow__"] = turbo_broccoli.tensorflow.from_json
        for t, f in DECODERS.items():
            if t in dct:
                return f(dct)
        return dct


class TurboBroccoliEncoder(json.JSONEncoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def default(self, o: Any) -> Any:

        ENCODERS: List[Callable[[Any], dict]] = [
            turbo_broccoli.bytes.to_json,
            turbo_broccoli.collections.to_json,
            turbo_broccoli.dataclass.to_json,
            turbo_broccoli.generic.to_json,
        ]
        if HAS_KERAS:
            ENCODERS.append(turbo_broccoli.keras.to_json)
        if HAS_NUMPY:
            ENCODERS.append(turbo_broccoli.numpy.to_json)
        if HAS_PANDAS:
            ENCODERS.append(turbo_broccoli.pandas.to_json)
        if HAS_TENSORFLOW:
            ENCODERS.append(turbo_broccoli.tensorflow.to_json)
        for f in ENCODERS:
            try:
                return f(o)
            except TypeError:
                pass
        return super().default(o)



"""Python standard collections (de)serialization"""
__docformat__ = "google"

from collections import deque
from typing import Any, Callable, List, Tuple


def _deque_to_json(deq: deque) -> dict:
    """Converts a deque into a JSON document."""
    return {
        "__type__": "deque",
        "__version__": 1,
        "data": list(deq),
        "maxlen": deq.maxlen,
    }


def _json_to_deque(dct: dict) -> deque:
    """
    Converts a JSON document to a deque. See `to_json` for the specification
    `dct` is expected to follow. Note that the key `__collections__` should not
    be present.
    """
    DECODERS = {
        1: _json_to_deque_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_deque_v1(dct: dict) -> Any:
    """Converts a JSON document to a deque following the v1 specification."""
    return deque(dct["data"], dct["maxlen"])


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a Python collection. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__collections__`.
    """
    DECODERS = {
        "deque": _json_to_deque,
    }
    try:
        return DECODERS[dct["__collections__"]["__type__"]](
            dct["__collections__"]
        )
    except KeyError as exc:
        raise TypeError("Not a valid collections document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a Python collection into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__collections__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    * `collections.deque`:

        {
            "__collections__": {
                "__type__": "deque,
                "__version__": 1,
                "data": [...],
                "maxlen": <int or None>,
            }
        }

    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (deque, _deque_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__collections__": f(obj)}
    raise TypeError("Not a supported collections type")
